
import gzip
from gzip import GzipFile
import logging
from pathlib import Path
import tarfile
from typing import Dict, Iterable, List, Tuple

from gencodegenes.transcript import Transcript

def find_uniprot_matches(path: Path, uniprot_ids: List[str]) -> Iterable[GzipFile]:
    ''' search a tar file for pdb files matching uniprot IDs
    '''
    logging.info(f'checking for structures in {path} with {uniprot_ids}')
    with tarfile.open(path) as tarred:
        for x in tarred:
            if x.name.endswith('cif.gz'):
                continue
            for uniprot_id in uniprot_ids:
                if uniprot_id in x.name:
                    logging.info(f'found structure for {uniprot_id}')
                    yield gzip.open(tarred.extractfile(x.name), 'rt')

def load_structure(path: Path, uniprot_ids: List[str]) -> Dict[Tuple[str, int], Dict[str, float]]:
    ''' a very basic PDB parser, to get positions of carbon atoms at the amino acids
    '''
    coords = None
    for handle in find_uniprot_matches(path, uniprot_ids):
        if coords is not None:
            # we've hit a second file for the same gene, so skip out
            return None
        
        coords = {}
        for line in handle:
            if not line.startswith('ATOM'):
                continue
            
            # only use the carbon atom position in each residue
            atom_name = line[12:16].strip(' ')
            if atom_name != 'C':
                continue
            
            residue = line[17:20]
            chain = line[21]
            residue_num = int(line[22:26])
            x_pos = float(line[30:38])
            y_pos = float(line[38:46])
            z_pos = float(line[46:54])
            
            key = (chain, residue_num)
            if key in coords:
                raise ValueError(f'seen this atom previously: {key}')
            coords[key] = {'residue': residue, 'x': x_pos, 'y': y_pos, 'z': z_pos}
        
        # TOOD: basic validation - do we have every residue included - check for gaps
    
    return coords

def valid_structure(structure_coords: Dict[Tuple[str, int], Dict[str, float]],
                    transcript: Transcript) -> bool:
    ''' check the structure sufficiently matches the transcript
    '''
    
    # check there is only a single protein chain in the structure, fail if not
    n_chains = len(set(x[0] for x in structure_coords))
    if n_chains > 1:
        logging.warning(f'cannot get distances from structure with multiple chains')
        return False
    
    # make sure coords are linearly sorted without gaps
    residues = [x[1] for x in sorted(structure_coords)]
    if residues != list(range(1, len(structure_coords) + 1)):
        logging.warning(f'cannot get distances from structure - missing residues')
        return False
    
    # make sure the protein coords extend to the end of the transcript range
    last_residue = transcript.get_coding_distance(transcript.cds_end)['pos'] // 3
    if abs(last_residue - residues[-1]) > 2:
        logging.warning(f"transcript length doesn't match structure length")
        return False
    
    return True


""" load trincleotide mutation rates
"""

from __future__ import annotations

import warnings
from pkg_resources import resource_filename

__PYSAM_ERROR = False
try:
    from pysam import VariantFile
except ImportError:
    __PYSAM_ERROR = True

def load_mutation_rates(path=None):
    """ load sequence context-based mutation rates
    
    Args:
        path: path to table of sequence context-based mutation rates. If None,
            this defaults to per-trinucleotide rates provided by Kaitlin Samocha
            (Broad Institute).
    
    Returns:
        list of [initial, changed, rate] lists e.g. [['AGA', 'ATA', '5e-8']]
    """
    
    if path is None:
        path = resource_filename(__name__, "data/rates.txt")
    
    rates = []
    with open(path) as handle:
        for line in handle:
            if line.startswith("from"): # ignore the header line
                continue
            
            line = [ x.encode('utf8') for x in line.strip().split() ]
            rates.append(line)
    
    return rates

class load_mutation_rates_in_region:
    ''' class to handle loading mutations rates within genome regions.

    This only permits loading data from files from Roulette e.g.
      http://genetics.bwh.harvard.edu/downloads/Vova/Roulette/
    '''
    CHROMS = list(map(str, range(1, 23)))
    def __init__(self, paths: list[str]):
        ''' start with list of paths to all Roulette VCFs
        '''
        if __PYSAM_ERROR:
            raise ValueError("pysam unavailable! try pip install denovonear[pysam] or pip install pysam")

        self.vcfs = {}
        for path in paths:
            tbx = VariantFile(path)
            # tag the tabixfile against all contigs (chromosomes) it contains
            for contig in self._check_chroms(tbx):
                if contig not in self.vcfs:
                    self.vcfs[contig] = []
                self.vcfs[contig].append(tbx)
    def _check_chroms(self, vcf):
        ''' find all the chromosomes actually present in a VCF
        '''
        contigs = []
        for chrom in self.CHROMS:
            try:
                next(vcf.fetch(chrom))
                contigs.append(chrom)
            except StopIteration:
                pass
        return contigs
    def __call__(self, chrom: str, start: int, end: int, tag='MR') -> dict[int, dict[str, float]]:
        ''' get mutation rates within a genome region

        Args:
            chrom: chromosome
            start: start position of region to gather data for
            end: end position (inclusive) of region to gather data for
            tag: which rate type to extract data for. Options are:
                MR (Roulette) - the default.
                GR (GnomAD) - from Kaitlin Samocha's rates
                AR (Adjusted roulette) - unavailable for most sites
                MC (Carlsson et al 2018) - unavailable for many sites.
        '''
        # the Roulette VCFs lack 'chr' prefixes, change to match
        if chrom.startswith('chr'):
            chrom = chrom[3:]
        
        # run through all the VCFs that contain data for the given chromosome.
        # Storing the data in a dict should deduplicate sites, if necessary.
        data = {}
        for tbx in self.vcfs[chrom]:
            for row in tbx.fetch(chrom, start, end):
                if row.pos not in data:
                    data[row.pos] = {}
                data[row.pos][row.alts[0]] = row.info[tag]
        return data




from math import log, isnan
from typing import Dict, List, Tuple, Union

from scipy.stats import chi2

from gencodegenes import Gene
from denovonear.load_gene import get_de_novos_in_transcript, minimise_transcripts
from denovonear.site_specific_rates import SiteRates
from denovonear.simulate import get_p_value, get_structure_p_value

def fishers_method(values):
    """ function to combine p values, using Fisher's method
    
    We occasionally have multiple P values for a mutation type, obtained from
    alternative transcripts for the gene. If we have only one P value for the
    gene for the mutation type, we just use that value, if we don't have any
    data, we use "NA", otherwise combine p-values using Fisher's method.
    
    Args:
        x: list of P-values for a gene
    
    Returns:
        combined P-value
    """
    values = [ x for x in values if not isnan(x) ]
    # use Fisher's combined method to estimate the P value from multiple
    # P-values. The chi square statistic is -2*sum(ln(P-values))
    return chi2.sf(-2 * sum(map(log, values)), 2 * len(values))

def cluster_de_novos(de_novos, gene, mut_dict, iterations=1000000):
    """ analysis proximity cluster of de novos in a single gene
    
    Args:
        de_novos: dictionary of de novo positions for the HGNC gene,
            indexed by functional type
        gene: gencodegenes.Gene object
        mut_dict: dictionary of mutation rates, indexed by trinuclotide sequence, 
        iterations: number of simulations to run
    
    Returns:
        a dictionary containing P values, and distances for missense, nonsense,
        and synonymous de novos events. Missing data is represented by "NA".
    """
    
    missense = de_novos["missense"]
    nonsense = de_novos["nonsense"]
    
    if len(gene.transcripts) == 0:
        nan = float('nan')
        return {'miss_dist': nan, 'miss_prob': nan, 'nons_prob': nan, 'nons_dist': nan}
    
    # load the set of transcripts that are the  minimum set of transcripts
    # required to contain all the de novos, unless we can't find any coding
    # transcripts that contain the de novos.
    transcripts = gene.transcripts
    minimized = minimise_transcripts(transcripts, missense + nonsense)
    transcripts = [x for x in transcripts if x.name in minimized]
    
    if len(transcripts) == 0:
        nan = float('nan')
        return {'miss_dist': nan, 'miss_prob': nan, 'nons_prob': nan, 'nons_dist': nan}
    
    probs = {"miss_prob": [], "nons_prob": []}
    dists = {"miss_dist": [], "nons_dist": []}
    
    for transcript in transcripts:
        
        missense_events = get_de_novos_in_transcript(transcript, missense)
        nonsense_events = get_de_novos_in_transcript(transcript, nonsense)
        
        # The rates either are passed in as lists of lists of sequence-context 
        # based rates which can be applied to a given sequence, or as 
        # dictionaries of rates per position/allele from genome based reference 
        # sources. If the latter, then we have to load the rates within the 
        # transcript region
        _rates = mut_dict
        if not isinstance(mut_dict, list):
            _rates = mut_dict(transcript.chrom, transcript.start, transcript.end)
        
        rates = SiteRates(transcript, _rates)
        
        (miss_dist, miss_prob) = get_p_value(transcript, rates, iterations, "missense", missense_events)
        (nons_dist, nons_prob) = get_p_value(transcript, rates, iterations, "lof", nonsense_events)
        
        dists["miss_dist"].append(miss_dist)
        dists["nons_dist"].append(nons_dist)
        probs["miss_prob"].append(miss_prob)
        probs["nons_prob"].append(nons_prob)
        
        # remove the de novos analysed in the current transcript, so that
        # analysis of subsequent transcripts uses independent events. NOTE THAT
        # THIS MIGHT MISS SOME CLUSTERING ACROSS MUTUALLY EXCLUSIVE TRANSCRIPTS
        # IF THE DE NOVO EVENTS ARE NEAR THE TRANSCRIPT DIVERGENCE.
        missense = [x for x in missense if x not in missense_events]
        nonsense = [x for x in nonsense if x not in  nonsense_events]
        
    for key in dists:
        dists[key] = ",".join([ str(x) for x in dists[key] ])
    
    probs = {k: fishers_method(probs[k]) for k in probs}
    probs.update(dists)
    
    return probs

def structure_cluster_de_novos(de_novos: Dict[str, List[int]],
                               structure: Dict[Tuple[str, int], Dict[str, float]],
                               gene: Gene,
                               mut_dict: Union[List[List[str]], Dict[int, Dict[str, float]]],
                               iterations=1000000,
                               ) -> Dict[str, float]:
    """ analysis proximity cluster of de novos in a single gene
    
    Args:
        de_novos: dictionary of de novo positions for the HGNC gene,
            indexed by functional type
        structure: 3D coordinates for carbon atoms in amino acids along protein 
            sequence, indexed by (chain, residue number)
        gene: gencodegenes.Gene object
        mut_dict: dictionary of mutation rates, indexed by trinuclotide sequence, 
        iterations: number of simulations to run
    
    Returns:
        a dictionary containing P values, and distances for missense and nonsense,
        de novos. Missing data is represented by "NA".
    """
    
    if len(gene.transcripts) == 0:
        nan = float('nan')
        return {'miss_dist': nan, 'miss_prob': nan, 'nons_prob': nan, 'nons_dist': nan}
    
    transcript = gene.canonical
    
    missense_events = get_de_novos_in_transcript(transcript, de_novos["missense"])
    nonsense_events = get_de_novos_in_transcript(transcript, de_novos["nonsense"])
    
    # The rates either are passed in as lists of lists of sequence-context 
    # based rates which can be applied to a given sequence, or as 
    # dictionaries of rates per position/allele from genome based reference 
    # sources. If the latter, then we have to load the rates within the 
    # transcript region
    _rates = mut_dict
    if not isinstance(mut_dict, list):
        _rates = mut_dict(transcript.chrom, transcript.start, transcript.end)
    
    rates = SiteRates(transcript, _rates)
    
    (miss_dist, miss_prob) = get_structure_p_value(transcript, rates, structure, iterations, "missense", missense_events)
    (nons_dist, nons_prob) = get_structure_p_value(transcript, rates, structure, iterations, "lof", nonsense_events)
    
    return {"miss_dist": miss_dist,
            "nons_dist": nons_dist,
            "miss_prob": miss_prob,
            "nons_prob": nons_prob,
    }
    
